# -*- coding: utf-8 -*-


def doc(func):
    """
        Find the message shown when someone calls the help command

    Parameters
    ----------
    func : function
        the function

    Returns
    -------
    str
        The help message for this command
    """
    stripped_chars = " \t"

    if hasattr(func, '__doc__'):
        docstring = func.__doc__.lstrip(" \n\t")
        if "\n" in docstring:
            i = docstring.index("\n")
            return docstring[:i].rstrip(stripped_chars)
        elif docstring:
            return docstring.rstrip(stripped_chars)

    return ""


def permission_check(data, command_permissions,
                     command=None, permissions=None):
    """
        Check the permissions of the user requesting a command

    Parameters
    ----------
    data : dict
        message data
    command_permissions : dict
        permissions of the command, contains all the roles as key and users
        with these permissions as values
    command : function
        the command that is run
    permissions : tuple or list
        a list of permissions for the command

    Returns
    -------
    bool
        True if the user has the right permissions, False otherwise
    """
    if permissions:
        pass
    elif command:
        if hasattr(command, 'permissions'):
            permissions = command.permissions
        else:
            return True  # true if no permission is required
    else:
        msg = "{name} must be called with command or permissions argument"
        raise RuntimeError(msg.format(name="_permission_check"))

    return any(data['sender']['id'] in command_permissions[permission]
               for permission in permissions
               if permission in command_permissions)


def restart_on(exc):
    """
        restart a function every time the `exc` exception is raised

    Parameters
    ----------
    exc : Exception
        the exception to catch
    """
    def decorator(func):
        while True:
            try:
                return func
            except exc:
                pass
            except:
                raise

    return decorator



# -*- coding: utf-8 -*-

import asyncio
import functools
import logging
import os
import sys

import peony

from . import exceptions

try:
    import magic
    mime = magic.Magic(mime=True)
except:  # pragma: no cover
    import mimetypes
    mime = mimetypes.MimeTypes()
    magic = None


_logger = logging.getLogger(__name__)


def error_handler(request):
    """
        The default error_handler

    The decorated request will retry infinitely on any handled error
    The exceptions handled are :class:`asyncio.TimeoutError` and
    :class:`exceptions.RateLimitExceeded`
    """

    @functools.wraps(request)
    async def decorated_request(url=None, **kwargs):
        tries = 3
        while True:
            try:
                return await request(url=url, **kwargs)

            except exceptions.RateLimitExceeded as e:
                delay = int(e.reset_in) + 1
                fmt = "Sleeping for {}s (rate limit exceeded on endpoint {})"
                _logger.warning(fmt.format(delay, url))
                await asyncio.sleep(delay)

            except asyncio.TimeoutError:
                fmt = "Request to {url} timed out, retrying"
                _logger.info(fmt.format(url=url))

            except exceptions.ServiceUnavailable:
                tries -= 1
                if not tries:
                    raise

                _logger.info("Service temporarily unavailable, "
                             "request will be made again very soon")
                await asyncio.sleep(0.5)

    return decorated_request


def get_args(func, skip=0):
    """
        Hackish way to get the arguments of a function

    Parameters
    ----------
    func : callable
        Function to get the arguments from
    skip : int, optional
        Arguments to skip, defaults to 0 set it to 1 to skip the
        ``self`` argument of a method.

    Returns
    -------
    tuple
        Function's arguments
    """

    code = getattr(func, '__code__', None)
    if code is None:
        code = func.__call__.__code__

    return code.co_varnames[skip:code.co_argcount]


def log_error(msg=None, exc_info=None, logger=None, **kwargs):
    """
        log an exception and its traceback on the logger defined

    Parameters
    ----------
    msg : str, optional
        A message to add to the error
    exc_info : tuple
        Information about the current exception
    logger : logging.Logger
        logger to use
    """
    if logger is None:
        logger = _logger

    if not exc_info:
        exc_info = sys.exc_info()

    if msg is None:
        msg = ""

    exc_class, exc_msg, _ = exc_info

    if all(info is not None for info in exc_info):
        logger.error(msg, exc_info=exc_info)


async def get_media_metadata(data, path=None):
    """
        Get all the file's metadata and read any kind of file object

    Parameters
    ----------
    data : bytes
        first bytes of the file (the mimetype shoudl be guessed from the
        file headers
    path : str, optional
        path to the file

    Returns
    -------
    str
        The mimetype of the media
    str
        The category of the media on Twitter
    """
    if isinstance(data, bytes):
        media_type = await get_type(data, path)

    else:
        raise TypeError("get_metadata input must be a bytes")

    media_category = get_category(media_type)

    _logger.info("media_type: %s, media_category: %s" % (media_type,
                                                         media_category))

    return media_type, media_category


async def get_size(media):
    """
        Get the size of a file

    Parameters
    ----------
    media : file object
        The file object of the media

    Returns
    -------
    int
        The size of the file
    """
    if hasattr(media, 'seek'):
        await execute(media.seek(0, os.SEEK_END))
        size = await execute(media.tell())
        await execute(media.seek(0))
    elif hasattr(media, 'headers'):
        size = int(media.headers['Content-Length'])
    elif isinstance(media, bytes):
        size = len(media)
    else:
        raise TypeError("Can't get size of media of type:",
                        type(media).__name__)

    _logger.info("media size: %dB" % size)
    return size


async def get_type(media, path=None):
    """
    Parameters
    ----------
    media : file object
        A file object of the image
    path : str, optional
        The path to the file

    Returns
    -------
    str
        The mimetype of the media
    str
        The category of the media on Twitter
    """
    if magic:
        if not media:
            raise TypeError("Media data is empty")

        _logger.debug("guessing mimetype using magic")
        media_type = mime.from_buffer(media[:1024])
    else:
        media_type = None
        if path:
            _logger.debug("guessing mimetype using built-in module")
            media_type = mime.guess_type(path)[0]

        if media_type is None:
            msg = ("Could not guess the mimetype of the media.\n"
                   "Please consider installing python-magic\n"
                   "(pip3 install peony-twitter[magic])")
            raise RuntimeError(msg)

    return media_type


def get_category(media_type):
    if media_type.startswith('video'):
        return "tweet_video"
    elif media_type.endswith('gif'):
        return "tweet_gif"
    elif media_type.startswith('image'):
        return "tweet_image"
    else:
        raise RuntimeError("The provided media cannot be handled.\n"
                           "mimetype: %s" % media_type)


async def execute(coro):
    """
        run a function or coroutine

    Parameters
    ----------
    coro : asyncio.coroutine or function
    """
    if asyncio.iscoroutine(coro):
        return await coro
    else:
        return coro


def set_debug():
    """ activates error messages, useful during development """
    logging.basicConfig(level=logging.WARNING)
    peony.logger.setLevel(logging.DEBUG)



# -*- coding: utf-8 -*-
"""
Peony Clients

:class:`BasePeonyClient` only handles requests while
:class:`PeonyClient` adds some methods that could help when using
the Twitter APIs, with a method to upload a media
"""

import asyncio
import io
import logging
from concurrent.futures import CancelledError
from urllib.parse import urlparse

import aiohttp

from . import data_processing, exceptions, general, oauth, utils
from .api import APIPath, StreamingAPIPath
from .commands import EventStreams, init_task, task
from .oauth import OAuth1Headers
from .stream import StreamResponse

logger = logging.getLogger(__name__)


class Setup(asyncio.Future):

    def __init__(self):
        super().__init__()
        self.early = asyncio.Future()


class MetaPeonyClient(type):

    def __new__(cls, name, bases, attrs, **kwargs):
        """ put the :class:`~peony.commands.tasks.Task`s in the right place """
        tasks = {'init_tasks': set(), 'tasks': set()}

        for base in bases:
            if hasattr(base, '_tasks'):
                for key, value in base._tasks.items():
                    tasks[key] |= value

        for attr in attrs.values():
            if isinstance(attr, init_task):
                tasks['init_tasks'].add(attr)
            elif isinstance(attr, task):
                tasks['tasks'].add(attr)

        attrs['_tasks'] = tasks

        return super().__new__(cls, name, bases, attrs)


class BasePeonyClient(metaclass=MetaPeonyClient):
    """
        Access the Twitter API easily

    You can create tasks by decorating a function from a child
    class with :class:`peony.task`

    You also attach a :class:`EventStream` to a subclass using
    the :func:`event_stream` of the subclass

    After creating an instance of the child class you will be able
    to run all the tasks easily by executing :func:`get_tasks`

    Parameters
    ----------
    streaming_apis : iterable, optional
        Iterable containing the streaming APIs subdomains
    base_url : str, optional
        Format of the url for all the requests
    api_version : str, optional
        Default API version
    suffix : str, optional
        Default suffix of API endpoints
    loads : function, optional
        Function used to load JSON data
    error_handler : function, optional
        Requests decorator
    session : aiohttp.ClientSession, optional
        Session to use to make requests
    proxy : str
        Proxy used with every request
    compression : bool, optional
        Activate data compression on every requests, defaults to True
    user_agent : str, optional
        Set a custom user agent header
    encoding : str, optional
        text encoding of the response from the server
    loop : event loop, optional
        An event loop, if not specified :func:`asyncio.get_event_loop`
        is called
    """

    _streams = EventStreams()

    def __init__(self,
                 consumer_key=None,
                 consumer_secret=None,
                 access_token=None,
                 access_token_secret=None,
                 bearer_token=None,
                 auth=None,
                 headers=None,
                 streaming_apis=None,
                 base_url=None,
                 api_version=None,
                 suffix='.json',
                 loads=data_processing.loads,
                 error_handler=utils.error_handler,
                 session=None,
                 proxy=None,
                 compression=True,
                 user_agent=None,
                 encoding=None,
                 loop=None,
                 **kwargs):

        if streaming_apis is None:
            self.streaming_apis = general.streaming_apis
        else:
            self.streaming_apis = streaming_apis

        if base_url is None:
            self.base_url = general.twitter_base_api_url
        else:
            self.base_url = base_url

        if api_version is None:
            self.api_version = general.twitter_api_version
        else:
            self.api_version = api_version

        if auth is None:
            auth = OAuth1Headers

        self.proxy = proxy

        self._suffix = suffix

        self.error_handler = error_handler

        self.encoding = encoding

        if encoding is not None:
            def _loads(*args, **kwargs):
                return loads(*args, encoding=encoding, **kwargs)

            self._loads = _loads
        else:
            self._loads = loads

        self.loop = asyncio.get_event_loop() if loop is None else loop

        self._session = session
        self._user_session = session is not None

        self._gathered_tasks = None

        if consumer_key is None or consumer_secret is None:
            raise TypeError("missing 2 required arguments: 'consumer_key' "
                            "and 'consumer_secret'")

        # all the possible args required by headers in :mod:`peony.oauth`
        kwargs = {
            'consumer_key': consumer_key,
            'consumer_secret': consumer_secret,
            'access_token': access_token,
            'access_token_secret': access_token_secret,
            'bearer_token': bearer_token,
            'compression': compression,
            'user_agent': user_agent,
            'headers': headers,
            'client': self
        }

        # get the args needed by the auth parameter on initialization
        args = utils.get_args(auth.__init__, skip=1)

        # keep only the arguments required by auth on init
        kwargs = {key: value for key, value in kwargs.items()
                  if key in args}

        self.headers = auth(**kwargs)

        self.setup = Setup()

        self.__setup = self.loop.create_task(self._setup())

    def init_tasks(self):
        """ tasks executed on initialization """
        return self._get_tasks(kind=init_task)

    async def _setup(self):
        if self._session is None:
            logger.debug("Creating session")
            self._session = aiohttp.ClientSession()

        # this will allow requests to be made starting from this point
        self.setup.early.set_result(True)

        init_tasks = self.init_tasks
        if callable(init_tasks):
            init_tasks = init_tasks()

        if init_tasks:
            logger.debug("Starting init tasks")
            await asyncio.gather(*init_tasks)

        self.setup.set_result(True)

    @staticmethod
    def _get_base_url(base_url, api, version):
        """
            create the base url for the api

        Parameters
        ----------
        base_url : str
            format of the base_url using {api} and {version}
        api : str
            name of the api to use
        version : str
            version of the api

        Returns
        -------
        str
            the base url of the api you want to use
        """
        format_args = {}

        if "{api}" in base_url:
            if api == "":
                base_url = base_url.replace('{api}.', '')
            else:
                format_args['api'] = api

        if "{version}" in base_url:
            if version == "":
                base_url = base_url.replace('/{version}', '')
            else:
                format_args['version'] = version

        return base_url.format(api=api, version=version)

    def __getitem__(self, values):
        """
            Access the api you want

        This permits the use of any API you could know about

        For most api you only need to type

        >>> self[api]  # api is the api you want to access

        You can specify a custom api version using the syntax

        >>> self[api, version]  # version is the api version as a str

        For more complex requests

        >>> self[api, version, suffix, base_url]

        Returns
        -------
        .api.BaseAPIPath
            To access an API endpoint
        """
        defaults = None, self.api_version, self._suffix, self.base_url
        keys = ['api', 'version', 'suffix', 'base_url']

        if isinstance(values, dict):
            # set values in the right order
            values = [values.get(key, defaults[i])
                      for i, key in enumerate(keys)]
        elif isinstance(values, set):
            raise TypeError('Cannot use a set to access an api, '
                            'please use a dict, a tuple or a list instead')
        elif isinstance(values, str):
            values = [values, *defaults[1:]]
        elif isinstance(values, tuple):
            if len(values) < len(keys):
                padding = (None,) * (len(keys) - len(values))
                values += padding

            values = [default if value is None else value
                      for value, default in zip(values, defaults)
                      if (value, default) != (None, None)]
        else:
            raise TypeError("Could not create an endpoint from an object of "
                            "type " + values.__class__.__name__)

        api, version, suffix, base_url = values

        base_url = self._get_base_url(base_url, api, version)

        # use StreamingAPIPath if subdomain is in self.streaming_apis
        if api in self.streaming_apis:
            return StreamingAPIPath([base_url], suffix=suffix, client=self)
        else:
            return APIPath([base_url], suffix=suffix, client=self)

    __getattr__ = __getitem__

    def __del__(self):
        if self.loop.is_closed():  # pragma: no cover
            pass
        elif self.loop.is_running():
            self.loop.create_task(self.close())
        else:
            self.loop.run_until_complete(self.close())

    async def request(self, method, url, future,
                      headers=None,
                      session=None,
                      encoding=None,
                      **kwargs):
        """
            Make requests to the REST API

        Parameters
        ----------
        future : asyncio.Future
            Future used to return the response
        method : str
            Method to be used by the request
        url : str
            URL of the resource
        headers : .oauth.PeonyHeaders
            Custom headers (doesn't overwrite `Authorization` headers)
        session : aiohttp.ClientSession, optional
            Client session used to make the request

        Returns
        -------
        data.PeonyResponse
            Response to the request
        """
        await self.setup.early

        # prepare request arguments, particularly the headers
        req_kwargs = await self.headers.prepare_request(
            method=method,
            url=url,
            headers=headers,
            **kwargs
        )

        if encoding is None:
            encoding = self.encoding

        if 'proxy' not in req_kwargs:
            req_kwargs['proxy'] = self.proxy

        session = session if (session is not None) else self._session

        logger.debug("making request with parameters: %s" % req_kwargs)

        async with session.request(**req_kwargs) as response:
            if response.status < 400:
                data = await data_processing.read(response, self._loads,
                                                  encoding=encoding)

                future.set_result(data_processing.PeonyResponse(
                    data=data,
                    headers=response.headers,
                    url=response.url,
                    request=req_kwargs
                ))
            else:  # throw exception if status is not 2xx
                await exceptions.throw(response, loads=self._loads,
                                       encoding=encoding, url=url)

    def stream_request(self, method, url, headers=None, _session=None,
                       *args, **kwargs):
        """
            Make requests to the Streaming API

        Parameters
        ----------
        method : str
            Method to be used by the request
        url : str
            URL of the resource
        headers : dict
            Custom headers (doesn't overwrite `Authorization` headers)
        _session : aiohttp.ClientSession, optional
            The session to use for this specific request, the session
            given as argument of :meth:`__init__` is used by default

        Returns
        -------
        .stream.StreamResponse
            Stream context for the request
        """
        return StreamResponse(
            method=method,
            url=url,
            client=self,
            headers=headers,
            session=_session,
            **kwargs
        )

    @classmethod
    def event_stream(cls, event_stream):
        """ Decorator to attach an event stream to the class """
        cls._streams.append(event_stream)
        return event_stream

    def _get_tasks(self, kind=task):
        if kind == task:
            key = 'tasks'
        elif kind == init_task:
            key = 'init_tasks'
        else:
            raise RuntimeError("Cannot get tasks of kind %s" % kind)

        return [task(self) for task in self._tasks[key]]

    def get_tasks(self):
        """
            Get the tasks attached to the instance

        Returns
        -------
        list
            List of tasks (:class:`asyncio.Task`)
        """
        tasks = self._get_tasks()

        tasks.extend(self._streams.get_tasks(self))

        return tasks

    async def run_tasks(self):
        """ Run the tasks attached to the instance """
        await self.setup
        tasks = self.get_tasks()
        self._gathered_tasks = asyncio.gather(*tasks, loop=self.loop)
        try:
            await self._gathered_tasks
        except CancelledError:
            pass

    async def arun(self):
        try:
            await self.run_tasks()
        except KeyboardInterrupt:
            pass
        finally:
            await self.close()

    def run(self):
        """ Run the tasks attached to the instance """
        self.loop.run_until_complete(self.arun())

    async def close(self):
        """ properly close the client """
        tasks = []
        # cancel setup
        if not self.setup.done():
            async def cancel_setup():
                self.__setup.cancel()
                try:
                    await self.__setup
                except:
                    pass

            tasks.append(cancel_setup())

        # close currently running tasks
        if self._gathered_tasks is not None:
            async def cancel_tasks():
                self._gathered_tasks.cancel()
                try:
                    await self._gathered_tasks
                except:
                    pass

            tasks.append(cancel_tasks())

        # close the session only if it was created by peony
        if not self._user_session:
            try:
                tasks.append(self._session.close())
            except (TypeError, AttributeError):
                pass

        if tasks:
            await asyncio.gather(*tasks)

        self._session = None


class PeonyClient(BasePeonyClient):
    """
        A client with some useful methods for most usages
    """

    def init_tasks(self):
        tasks = super().init_tasks()
        if isinstance(self.headers, oauth.OAuth1Headers):
            tasks.append(self.__get_user())
        return tasks

    @init_task
    async def __get_twitter_configuration(self):
        """
        create a ``twitter_configuration`` attribute with the response
        of the endpoint
        https://api.twitter.com/1.1/help/configuration.json
        """
        api = self['api', general.twitter_api_version,
                   ".json", general.twitter_base_api_url]

        req = api.help.configuration.get()
        self.twitter_configuration = await req

    async def __get_user(self):
        """
        create a ``user`` attribute with the response of the endpoint
        https://api.twitter.com/1.1/account/verify_credentials.json
        """
        api = self['api', general.twitter_api_version,
                   ".json", general.twitter_base_api_url]

        req = api.account.verify_credentials.get()
        self.user = await req

    async def _chunked_upload(self, media, media_size,
                              path=None,
                              media_type=None,
                              media_category=None,
                              chunk_size=2**20,
                              **params):
        """
            upload media in chunks

        Parameters
        ----------
        media : file object
            a file object of the media
        media_size : int
            size of the media
        path : str, optional
            filename of the media
        media_type : str, optional
            mime type of the media
        media_category : str, optional
            twitter media category, must be used with ``media_type``
        chunk_size : int, optional
            size of a chunk in bytes
        params : dict, optional
            additional parameters of the request

        Returns
        -------
        .data_processing.PeonyResponse
            Response of the request
        """
        if isinstance(media, bytes):
            media = io.BytesIO(media)

        chunk = media.read(chunk_size)
        is_coro = asyncio.iscoroutine(chunk)

        if is_coro:
            chunk = await chunk

        if media_type is None:
            media_metadata = await utils.get_media_metadata(chunk, path)
            media_type, media_category = media_metadata
        elif media_category is None:
            media_category = utils.get_category(media_type)

        response = await self.upload.media.upload.post(
            command="INIT",
            total_bytes=media_size,
            media_type=media_type,
            media_category=media_category,
            **params
        )

        media_id = response['media_id']
        i = 0

        while chunk:
            if is_coro:
                req = self.upload.media.upload.post(command="APPEND",
                                                    media_id=media_id,
                                                    media=chunk,
                                                    segment_index=i)
                chunk, _ = await asyncio.gather(media.read(chunk_size), req)
            else:
                await self.upload.media.upload.post(command="APPEND",
                                                    media_id=media_id,
                                                    media=chunk,
                                                    segment_index=i)

                chunk = media.read(chunk_size)

            i += 1

        status = await self.upload.media.upload.post(command="FINALIZE",
                                                     media_id=media_id)

        if 'processing_info' in status:
            while status['processing_info'].get('state') != "succeeded":
                processing_info = status['processing_info']
                if processing_info.get('state') == "failed":
                    error = processing_info.get('error', {})

                    message = error.get('message', str(status))

                    raise exceptions.MediaProcessingError(data=status,
                                                          message=message,
                                                          **params)

                delay = processing_info['check_after_secs']
                await asyncio.sleep(delay)

                status = await self.upload.media.upload.get(
                    command="STATUS",
                    media_id=media_id,
                    **params
                )

        return response

    def _size_test(self, size, size_limit):
        if size_limit is None:
            if isinstance(self.twitter_configuration, APIPath):
                return False
            else:
                try:
                    size_limit = self.twitter_configuration['photo_size_limit']
                except (KeyError, TypeError):
                    return False

        return size > size_limit

    async def upload_media(self, file_,
                           media_type=None,
                           media_category=None,
                           chunked=None,
                           size_limit=None,
                           **params):
        """
            upload a media on twitter

        Parameters
        ----------
        file_ : str or pathlib.Path or file
            Path to the file or file object
        media_type : str, optional
            mime type of the media
        media_category : str, optional
            Twitter's media category of the media, must be used with
            ``media_type``
        chunked : bool, optional
            If True, force the use of the chunked upload for the media
        size_limit : int, optional
            If set, the media will be sent using a multipart upload if
            its size is over ``size_limit`` bytes
        params : dict
            parameters used when making the request

        Returns
        -------
        .data_processing.PeonyResponse
            Response of the request
        """
        if isinstance(file_, str):
            url = urlparse(file_)
            if url.scheme.startswith('http'):
                media = await self._session.get(file_)
            else:
                path = urlparse(file_).path.strip(" \"'")
                media = await utils.execute(open(path, 'rb'))
        elif hasattr(file_, 'read') or isinstance(file_, bytes):
            media = file_
        else:
            raise TypeError("upload_media input must be a file object or a "
                            "filename or binary data or an aiohttp request")

        media_size = await utils.get_size(media)
        size_test = self._size_test(media_size, size_limit)

        if isinstance(media, aiohttp.ClientResponse):
            # send the content of the response
            media = media.content

        if (size_test and chunked is None) or chunked:
            args = media, media_size, file_, media_type, media_category
            response = await self._chunked_upload(*args, **params)
        else:
            response = await self.upload.media.upload.post(media=media,
                                                           **params)

        if not hasattr(file_, 'read') and not getattr(media, 'closed', True):
            media.close()

        return response



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod

from . import general, requests


class AbstractAPIPath(ABC):
    """
        The syntactic sugar factory

    Every time you get an attribute or an item from an instance of this
    class this will be appended to its ``_path`` until you call a request
    method (like get or post)

    It makes it easy to call any endpoint of the api

    The ``client`` given as an parameter during the creation of the
    BaseAPIPath instance can be accessed as the ``_client`` attribute of
    the instance.

    .. warning::

        You must create a child class of AbstractAPIPath to perform
        requests (you have to implement the _request method)

    Parameters
    ----------
    path : str
        Value of ``_path``
    suffix : str
        suffix to append to the url
    client : .client.BasePeonyClient
        client used to perform the request

    """

    def __init__(self, path, suffix, client):
        self._path = path
        self._suffix = suffix
        self._client = client

    def url(self, suffix=None):
        """
            Build the url using the _path attribute

        Parameters
        ----------
        suffix : str
            String to be appended to the url

        Returns
        -------
        str
            Path to the endpoint
        """
        return "/".join(self._path) + (suffix or self._suffix)

    def __getitem__(self, key):
        """
            Where the magic happens

        If the key is a request method (eg. get) call the _request
        attribute with the method as argument

        otherwise append the key to the _path attribute

        >>> api = APIPath()  # you would have to add more arguments
        >>> api['_client']   # appends '_client' to _path

        Parameters
        ----------
        key : :obj:`str`, :obj:`tuple` or :obj:`list`
            Key used to access an API endpoint and appended to the
            path attribute

        Returns
        -------
        BaseAPIPath
            New APIPath instance with a new ``path`` value
        """
        if isinstance(key, str):
            if key.lower() in general.request_methods:
                return self._request(key)
            else:
                new_path = self._path + [key]
        elif isinstance(key, (tuple, list)):
            key = [str(i) for i in key]
            new_path = self._path + key
        else:
            raise TypeError("Could not create endpoint from %s "
                            "of type %s" % (key, type(key)))

        return self.__class__(path=new_path,
                              suffix=self._suffix,
                              client=self._client)

    def __getattr__(self, k):
        """
            Call __getitem__ when trying to get an attribute from the
            instance

        If your path contains an actual attribute of the instance
        you should call __getitem__ instead
        """
        return self[k]

    @abstractmethod
    def _request(self, method):
        """
            Make a request for the endpoint

        Parameters
        ----------
        method : str
            method to use to make the request
        """

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.url())


class APIPath(AbstractAPIPath):
    """
        Class to make requests to a REST API

    Parameters
    ----------
    path : str
        Value of ``_path``
    suffix : str
        suffix to append to the url
    client : .client.BasePeonyClient
        client used to perform the request
    """

    def _request(self, method):
        return requests.RequestFactory(self, method)


class StreamingAPIPath(AbstractAPIPath):
    """
        Class to make requests to a Streaming API

    Parameters
    ----------
    path : str
        Value of ``_path``
    suffix : str
        suffix to append to the url
    client : .client.BasePeonyClient
        client used to perform the request
    """

    def _request(self, method):
        return requests.StreamingRequest(self, method)



# -*- coding: utf-8 -*-
"""
===============
 peony-twitter
===============

  An asynchronous Twitter API client for Python 3.5+

"""

import logging

__author__ = "Florian Badie"
__author_email__ = "florianbadie@gmail.com"
__url__ = "https://github.com/odrling/peony-twitter"

__version__ = "1.0.4"

__license__ = "MIT License"

__keywords__ = "twitter, asyncio, asynchronous"

logger = logging.getLogger(__name__)

from .client import BasePeonyClient, PeonyClient  # noqa
from .commands import EventStream, event_handler, events, task, init_task  # noqa
from .utils import set_debug  # noqa



# -*- coding: utf-8 -*-

import asyncio
import sys
from abc import ABC, abstractmethod


class AbstractIterator(ABC):
    """
        Asynchronous iterator

    Parameters
    ----------
    request : .requests.Request
        Main request
    """

    def __init__(self, request):
        self.request = request
        self.kwargs = request.kwargs.copy()

    def __aiter__(self):
        return self

    if sys.version_info < (3, 5, 2):  # pragma: no cover
        __aiter__ = asyncio.coroutine(__aiter__)

    @abstractmethod
    async def __anext__(self):
        """ the function called on each iteration """


class IdIterator(AbstractIterator):
    """
        Iterate using ids

    It is the parent class of MaxIdIterator and SinceIdIterator

    Parameters
    ----------
    request : .requests.Request
        Main request
    parameter : str
        Parameter to change for each request
    force : bool
        Keep the iterator after empty responses
    """

    def __init__(self, request, parameter, force=False):
        """ Keep all the arguments as class attributes """
        self.param = parameter
        self.force = force
        super().__init__(request)

    async def __anext__(self):
        """ return each response until getting an empty data """
        request = self.request(**self.kwargs)
        response = await request
        if response:
            response = await self.call_on_response(response)
        elif not self.force:
            raise StopAsyncIteration

        return response

    @abstractmethod
    async def call_on_response(self, response):
        """ function that prepares for the next request """


class MaxIdIterator(IdIterator):
    """
        Iterator for endpoints using max_id

    Parameters
    ----------
    request : .requests.Request
        Main request
    """

    def __init__(self, request):
        super().__init__(request,
                         parameter="max_id",
                         force=False)

    async def call_on_response(self, response):
        """
            The parameter is set to the id of the tweet at index i - 1
        """
        self.kwargs[self.param] = response[-1]['id'] - 1
        return response


class SinceIdIterator(IdIterator):
    """
        Iterator for endpoints using since_id

    Parameters
    ----------
    request : .requests.Request
        Main request
    force : bool
        Keep the iterator after empty responses
    fill_gaps : bool
        Fill the gaps (if there are more than ``count`` tweets to get)
    """

    def __init__(self, request, force=True, fill_gaps=False):
        super().__init__(request,
                         parameter="since_id",
                         force=force)

        self.fill_gaps = fill_gaps
        self.last_id = None

    async def set_param(self, response):
        if response:
            if self.fill_gaps:
                self.kwargs[self.param] = response[0]['id'] - 1
                self.last_id = response[0]['id']
            else:
                self.kwargs[self.param] = response[0]['id']

    async def call_on_response(self, response):
        """
        Try to fill the gaps and strip last tweet from the response
        if its id is that of the first tweet of the last response

        Parameters
        ----------
        response : dict or list
            The response
        """
        since_id = self.kwargs.get(self.param, 0) + 1

        if self.fill_gaps:
            if response[-1]['id'] != since_id:
                max_id = response[-1]['id'] - 1
                responses = with_max_id(self.request(**self.kwargs,
                                                     max_id=max_id))

                async for tweets in responses:
                    response.extend(tweets)

            if response[-1]['id'] == self.last_id:
                response = response[:-1]

                if not response and not self.force:
                    raise StopAsyncIteration

        await self.set_param(response)

        return response


class CursorIterator(AbstractIterator):
    """
        Iterate using a cursor

    Parameters
    ----------
    request : .requests.Request
        Main request
    """

    async def __anext__(self):
        """ return each response until getting 0 as next cursor """
        if self.kwargs.get('cursor', -1) != 0:
            response = await self.request(**self.kwargs)
            self.kwargs['cursor'] = response['next_cursor']
            return response
        else:
            raise StopAsyncIteration


with_max_id = MaxIdIterator
with_since_id = SinceIdIterator
with_cursor = CursorIterator



# -*- coding: utf-8 -*-

import asyncio
from abc import ABC, abstractmethod
from types import GeneratorType

from . import iterators, utils

iterable = (list, set, tuple, GeneratorType)


class Endpoint:
    """
        A class representing an endpoint

    Parameters
    ----------
    api : api.AbstractAPIPath
        API path of the request
    method : str
        HTTP method to be used by the request
    """

    def __init__(self, *request):
        if len(request) == 1:
            request = request[0]
            self.api = request.api
            self.method = request.method
        else:
            self.api, self.method = request


class AbstractRequest(ABC):
    """
        A function that makes a request when called
    """

    def _get_params(self, _suffix=None, **kwargs):
        if _suffix is None:
            _suffix = self.api._suffix

        kwargs, skip_params = self.sanitize_params(self.method, **kwargs)

        return kwargs, skip_params, self.api.url(_suffix)

    @staticmethod
    def sanitize_params(method, **kwargs):
        """
            Request params can be extracted from the ``**kwargs``

        Arguments starting with `_` will be stripped from it, so they
        can be used as an argument for the request
        (eg. "_headers" → "headers" in the kwargs returned by this
        function while "headers" would be inserted into the parameters
        of the request)

        Parameters
        ----------
        method : str
            method to use to make the request
        kwargs : dict
            Keywords arguments given to the request

        Returns
        -------
        dict
            New requests parameters, correctly formatted
        """
        # items which does not have a key starting with `_`
        items = [(key, value) for key, value in kwargs.items()
                 if not key.startswith("_")]
        params, skip_params = {}, False

        for key, value in items:
            # binary data
            if hasattr(value, 'read') or isinstance(value, bytes):
                params[key] = value
                # The params won't be used to make the signature
                skip_params = True

            # booleans conversion
            elif isinstance(value, bool):
                params[key] = "true" if value else "false"

            # iterables conversion
            elif isinstance(value, iterable):
                params[key] = ",".join(map(str, value))

            # skip params with value None
            elif value is None:
                pass

            # the rest is converted to str
            # (make sure you don't send something wrong)
            else:
                params[key] = str(value)

        # dict with other items (+ strip "_" from keys)
        kwargs = {key[1:]: value for key, value in kwargs.items()
                  if key.startswith("_")}

        if method == "post" and not kwargs.get('data', None) and params:
            kwargs['data'] = params  # post requests use the data argument

        elif not kwargs.get('params', None) and params:
            kwargs['params'] = params

        return kwargs, skip_params

    @abstractmethod
    def __call__(self, **kwargs):
        """ method called to make the request """


class Iterators(Endpoint):
    """
    Access the iterators from :mod:`peony.iterators` right from a
    request object
    """

    def __init__(self, request):
        super().__init__(request)
        self.request = request

    def __getattr__(self, key):
        iterator = getattr(iterators, key)

        if isinstance(self.request, Request):
            def iterate(**kwargs):
                return iterator(self.request, **kwargs)
        else:
            keys = utils.get_args(iterator.__init__)

            def iterate(**kwargs):
                iterator_kwargs = {}
                for key in keys:
                    if '_' + key in kwargs:
                        iterator_kwargs[key] = kwargs.pop('_' + key)

                request = self.request(**kwargs)

                return iterator(request, **iterator_kwargs)

        return iterate


class RequestFactory(Endpoint):
    """
        Requests to REST APIs

    Parameters
    ----------
    api : api.AbstractAPIPath
        API path of the request
    method : str
        HTTP method to be used by the request
    """

    def __init__(self, api, method):
        super().__init__(api, method)
        self.iterator = Iterators(self)

    def __call__(self, **kwargs):
        return Request(self.api, self.method, **kwargs)


class Request(asyncio.Future, AbstractRequest):

    def __init__(self, api, method, **kwargs):
        super().__init__()
        self.api = api
        self.method = method
        self.iterator = Iterators(self)
        self.kwargs = kwargs

        kwargs, skip_params, url = self._get_params(**kwargs)

        # if user explicitly wants to skip parameters in the oauth signature
        if 'skip_params' in kwargs:
            skip_params = kwargs.pop('skip_params')

        error_handling = kwargs.pop('error_handling', True)

        kwargs.update(method=self.method, url=url, skip_params=skip_params)

        client = self.api._client
        request = client.request

        if client.error_handler and error_handling:
            request = self.api._client.error_handler(request)

        client.loop.create_task(request(future=self, **kwargs))

    def __call__(self, **kwargs):
        return self.__class__(self.api, self.method, **kwargs)


class StreamingRequest(AbstractRequest):
    """
        Requests to Streaming APIs
    """

    def __init__(self, api, method):
        self.api = api
        self.method = method

    def __call__(self, **kwargs):
        kwargs, skip_params, url = self._get_params(**kwargs)

        return self.api._client.stream_request(self.method, url=url,
                                               skip_params=skip_params,
                                               **kwargs)



# -*- coding: utf-8 -*-

from functools import wraps

from ..utils import get_args
from .event_handlers import EventHandler

on = 'on_{name}'


def _get_value(func):
    value = func()

    if value is None:
        value = func.__name__

    return value


class Handler:
    """
        A decorator, the decorated function is used when the
        event is detected related to this handler is detected

    Parameters
    ----------
    event : func
        a function that returns True when the data received
        corresponds to an event
    """

    def __init__(self, event):
        self.event = event

    def __call__(self, func):
        return EventHandler(func=func, event=self.event)

    def with_prefix(self, prefix, strict=False):
        """
            decorator to handle commands with prefixes

        Parameters
        ----------
        prefix : str
            the prefix of the command
        strict : bool, optional
            If set to True the command must be at the beginning
            of the message. Defaults to False.

        Returns
        -------
        function
            a decorator that returns an :class:`EventHandler` instance
        """

        def decorated(func):
            return EventHandler(func=func, event=self.event,
                                prefix=prefix, strict=strict)

        return decorated


class Event:
    """
        Represents an event, the handler attribute is
        an instance of Handler

    Parameters
    ----------
    func : callable
        a function that returns True when the data received
        corresponds to an event
    name : str
        name given to the event
    """

    def __init__(self, func, name):
        self._func = func
        self.handler = Handler(func)
        self.__name__ = name
        self.__doc__ = func.__doc__

    def envelope(self):
        """ returns an :class:`Event` that can be used for site streams """

        def enveloped_event(data):
            return 'for_user' in data and self._func(data.get('message'))

        return self.__class__(enveloped_event, self.__name__)

    for_user = envelope

    def __call__(self, data):
        return self._func(data)

    def __str__(self):
        return "Event {name}".format(name=self.__name__)

    def __repr__(self):
        return str(self)


class Events(dict):
    """
        A class to manage event handlers easily
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aliases = {}

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, func):
        event = func if isinstance(func, Event) else Event(func, key)
        setattr(self, key, event)

    def _set_aliases(self, *keys, event=None, func=None):
        keys = [key if "{name}" not in key else key.format(name=func.__name__)
                for key in keys]

        if func:
            event = self(func)
        elif event:
            event = self.event(event)
        else:
            raise RuntimeError("Could not set alias")

        for key in keys:
            self[key] = event
            self.aliases[key] = self[func.__name__]

        return event

    def alias(self, *keys):

        def decorator(func):
            func.__doc__ += "\n:aliases: %s\n" % ', '.join(keys)
            event = self._set_aliases(*keys, func=func)

            return event

        return decorator

    def event_alias(self, *keys):

        def decorator(func):
            func.__doc__ += "\n:aliases: %s\n" % ', '.join(keys)
            event = self._set_aliases(*keys, event=func)

            return event

        return decorator

    def event(self, func):
        if not len(get_args(func)):
            value = _get_value(func)

            @wraps(func)
            def decorated(data):
                return data.get('event') == value

            self[func.__name__] = decorated
            self['on_' + func.__name__] = decorated
            self[func.__name__].__doc__ = func.__doc__
            return self[func.__name__]
        else:
            self[func.__name__] = func
            return self[func.__name__]

    def __call__(self, func):
        if not len(get_args(func)):
            value = _get_value(func)

            @wraps(func)
            def decorated(data):
                return value in data

            self[func.__name__] = decorated
        else:
            self[func.__name__] = func

        return self[func.__name__]

    @property
    def no_aliases(self):
        return {key: value for key, value in self.items()
                if key not in self.aliases}

    def priority(self, p):

        def decorated(func):
            func.priority = p
            return self(func)

        return decorated


events = Events()


@events
def friends(data):
    """
        Event triggered on connection to an userstream

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#friends-lists-friends
    """  # noqa: E501
    return 'friends' in data or 'friends_str' in data


@events.alias(on, 'on_dm')
def direct_message():
    """
        Event triggered when a direct message is received

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#direct-messages
    """


@events.alias(on)
@events.priority(-1)
def retweeted_status(data):
    """
        Event triggered when the data corresponds to a retweet

    For more information:
    https://dev.twitter.com/overview/api/tweets
    """
    return tweet(data) and 'retweeted_status' in data


@events.alias(on)
def tweet(data):
    """
        Event triggered when the data corresponds to a tweet
    If there is no handler for the :func:`retweeted_status` event
    then the data could correspond to a retweet

    For more information:
    https://dev.twitter.com/overview/api/tweets
    """
    return 'text' in data


@events.alias(on, 'deleted_tweet')
def delete():
    """
        Event triggered when an user deletes a tweet

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#status-deletion-notices-delete
    """  # noqa: E501


@events.alias('location_deleted')
def scrub_geo():
    """
        Event triggered when an user deletes their location on a range of
        tweets

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#location-deletion-notices-scrub-geo
    """  # noqa: E501


@events.event
def limit():
    """
        Event triggered when the data corresponds to a stream limit notice

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#limit-notices-limit
    """  # noqa: E501


@events.event
def status_withheld():
    """
        Event triggered upon receiving a status withheld notice

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#withheld-content-notices-status-withheld-user-withheld
    """  # noqa: E501


@events.event
def user_withheld():
    """
        Event triggered upon receiving a status withheld notice

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#user-withheld
    """


@events.alias(on)
def disconnect():
    """
        Event triggered upon receiving a disconnect notice
    Note that the disconnect message may not be received when experiencing
    network issues.

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#disconnect-messages-disconnect
    """  # noqa: E501


# warnings

@events.alias(on)
def warning():
    """
        Event triggered when receiving a warning

    For more information:

    * https://dev.twitter.com/streaming/overview/messages-types#stall-warnings-warning
    * https://dev.twitter.com/streaming/overview/messages-types#too-many-follows-warning
    """  # noqa: E501


@events.alias(on)
def stall_warning(data):
    """
        Event triggered when receiving a stall warning

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#stall-warnings-warning
    """  # noqa: E501
    return (warning(data) and
            data.get('warning').get('code') == "FALLING_BEHIND")


@events.alias(on)
def too_many_follows(data):
    """
        Event triggered when receiving a "too many follows" warning

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#too-many-follows-warning
    """  # noqa: E501
    return (warning(data) and
            data.get('warning').get('code') == "FOLLOWS_OVER_LIMIT")

# events, the data looks like; {"event": EVENT_NAME, ...}


@events.event
def access_revoked():
    """
        Event triggered when the user is deauthorized

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def follow():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def unfollow():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def block():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def unblock():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def favorite():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def unfavorite():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def list_created():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def list_destroyed():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def list_updated():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def list_member_added():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def list_member_removed():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def list_user_subscribed():  # noqa: E501
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def list_user_unsubscribed():  # noqa: E501
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def quoted_tweet():
    """

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#events-event
    """


@events.event
def user_update():
    """
            Event triggered when an user updates their profile

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#user_update
    """


# Site stream control messages

@events.alias('control_message')
def control():
    """
        Event triggered upon receiving a control message

    For more information:
    https://dev.twitter.com/streaming/overview/messages-types#control-messages-control
    """  # noqa: E501


# Internal peony events

@events.alias(on, 'first_connection')
def connected():
    """
        event_triggered on the first connection to a stream
    """


@events.alias(on)
@events.priority(1)
def connect(data):
    """
        event triggered on connection or reconnection to a stream
    """
    return connected(data) or stream_restart(data)


@events.alias(on, 'on_restart', 'restart')
def stream_restart():
    """
        Event triggered on stream restart
    """


@events.alias(on, 'reconnect', 'on_reconnect')
def reconnecting_in():
    """
        Event triggered when a stream restart is scheduled

    the data contains the number of seconds to wait is seconds as its
    ``'reconnecting_in'`` item.
    """


# matches any event that wasn't handled
@events.priority(1024)
def default(_):
    """
        Event triggered when the data didn't trigger any handled event
    """
    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import inspect
import os
import re
import sys

from setuptools import find_packages, setup


def get_metadata(metadata_file):
    with open(metadata_file) as stream:
        text = stream.read()

        ex = r"__(\w*?)__\s*?=\s*?[\"\']([^\"\']*)"
        items = re.findall(ex, text)
        metadata = {key: value for key, value in items}

        metadata['keywords'] = metadata.get('keywords', '').split(', ')

        ex = r'"{3}[^\w]*(?P<name>[^\s]*)[^\w]+(?P<description>.*)'
        match = re.search(ex, text)

        metadata.update(match.groupdict())

    return metadata


ex = r"# ([^\s:]+).*\n([^#]+)"
requirements_prog = re.compile(ex)


def get_requirements(requirements):
    with open(requirements) as stream:
        matches = requirements_prog.findall(stream.read())
        requires = {key: value.strip().splitlines()
                    for key, value in matches}

    return requires


def main():
    if sys.version_info < (3, 5):
        raise RuntimeError("Peony requires Python 3.5+")

    dirname = os.path.dirname(inspect.getfile(inspect.currentframe()))

    # get metadata and keywords from peony/__init__.py
    kwargs = get_metadata(os.path.join(dirname, 'peony', '__init__.py'))

    # get requirements from requirements.txt
    kwargs.update(get_requirements(os.path.join(dirname, 'requirements.txt')))

    # get extras requirements from extras_require.txt
    extras = os.path.join(dirname, 'extras_require.txt')
    extras_require = get_requirements(extras)

    # get long description from README.md
    with open('README.rst') as stream:
        long_description = stream.read()

    setup(long_description=long_description,
          packages=find_packages(include=["peony*"]),
          extras_require=extras_require,
          python_requires='>=3.5.0',
          **kwargs)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

import asyncio
import logging
import sys
from concurrent.futures import CancelledError

import aiohttp
import async_timeout

from . import data_processing, exceptions, utils
from .exceptions import StreamLimit
from .general import rate_limit_notices

ClientPayloadError = aiohttp.ClientPayloadError
ClientConnectionError = aiohttp.ClientConnectionError

RECONNECTION_TIMEOUT = 5
MAX_RECONNECTION_TIMEOUT = 320
DISCONNECTION_TIMEOUT = 0.25
ERROR_TIMEOUT = DISCONNECTION_TIMEOUT
MAX_DISCONNECTION_TIMEOUT = 16
ENHANCE_YOUR_CALM_TIMEOUT = 60

NORMAL = 0
DISCONNECTION = 1
ERROR = DISCONNECTION
RECONNECTION = 2
ENHANCE_YOUR_CALM = 3

HandledErrors = asyncio.TimeoutError, ClientPayloadError, TimeoutError

logger = logging.getLogger(__name__)


class StreamResponse:
    """
        Asynchronous iterator for streams

    Parameters
    ----------
    *args : list, optional
        Positional arguments of the request
    client : .client.BasePeonyClient
        client used to make the request
    session : aiohttp.ClientSession, optional
        Session used by the request
    loads : function, optional
        function used to decode the JSON data received
    timeout : int, optional
        Timeout on connection
    kwargs : dict, optional
        Keyword parameters of the request
    """

    def __init__(self, *args,
                 client,
                 session=None,
                 loads=data_processing.loads,
                 timeout=10,
                 **kwargs):

        self.client = client
        self.session = session
        self.loads = loads
        self.timeout = timeout
        self.args = args
        self.kwargs = kwargs

        self.response = None
        self._reconnecting = False
        self._state = NORMAL
        self._error_timeout = 0

    async def _connect(self):
        """
            Connect to the stream

        Returns
        -------
        asyncio.coroutine
            The streaming response
        """
        logger.debug("connecting to the stream")
        await self.client.setup.early
        if self.session is None:
            self.session = self.client._session
        kwargs = await self.client.headers.prepare_request(**self.kwargs)
        request = self.client.error_handler(self.session.request)

        return await request(*self.args, timeout=0, **kwargs)

    async def connect(self):
        """
            Create the connection

        Returns
        -------
        self

        Raises
        ------
        exception.PeonyException
            On a response status in 4xx that are not status 420 or 429
            Also on statuses in 1xx or 3xx since this should not be the status
            received here
        """
        with async_timeout.timeout(self.timeout):
            self.response = await self._connect()

        if self.response.status in range(200, 300):
            self._error_timeout = 0
            self.state = NORMAL
        elif self.response.status == 500:
            self.state = DISCONNECTION
        elif self.response.status in range(501, 600):
            self.state = RECONNECTION
        elif self.response.status in (420, 429):
            self.state = ENHANCE_YOUR_CALM
        else:
            logger.debug("raising error during stream connection")
            raise await exceptions.throw(self.response,
                                         loads=self.client._loads,
                                         url=self.kwargs['url'])

        logger.debug("stream state: %d" % self.state)

    def __aiter__(self):
        return self

    if sys.version_info < (3, 5, 2):  # pragma: no cover
        __aiter__ = asyncio.coroutine(__aiter__)

    async def __anext__(self):
        """
            Decode each line using json

        Returns
        -------
        dict
            Decoded JSON data
        """
        if self.response is None:
            logger.info("first connection to the stream")
            await self.connect()
            return {'connected': True}

        line = b''
        try:
            if self.state != NORMAL:
                if self._reconnecting:
                    return await self.restart_stream()
                else:
                    return await self.init_restart()

            while not line:
                with async_timeout.timeout(90):
                    line = await self.response.content.readline()
                    line = line.strip(b'\r\n')
                    logger.debug("received data: %s" % line)

            if line in rate_limit_notices:
                logger.debug("raising StreamLimit")
                raise StreamLimit(line)

            logger.debug("decoding data")
            return self.loads(line)

        except HandledErrors as e:
            logger.debug("handling error %s: %s" % (e.__class__.__name__, e))
            self.state = ERROR
            return await self.init_restart()

        except ClientConnectionError:
            logger.debug("Disconnected from stream")
            self.state = DISCONNECTION
            return await self.init_restart()

        except CancelledError:
            logger.debug("Stopping stream")
            raise StopAsyncIteration

        except Exception as e:
            self.state = ERROR
            return await self.init_restart(error=e)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        if value == NORMAL or self.state < value:
            self._state = value

    async def init_restart(self, error=None):
        """
            Restart the stream on error

        Parameters
        ----------
        error : bool, optional
            Whether to print the error or not
        """
        if error:
            utils.log_error(logger=logger)

        if self.state == DISCONNECTION:
            if self._error_timeout < MAX_DISCONNECTION_TIMEOUT:
                self._error_timeout += DISCONNECTION_TIMEOUT

            logger.info("The stream was disconnected, will reconnect in %ss"
                        % self._error_timeout)

        elif self.state == RECONNECTION:
            if self._error_timeout < RECONNECTION_TIMEOUT:
                self._error_timeout = RECONNECTION_TIMEOUT
            elif self._error_timeout < MAX_RECONNECTION_TIMEOUT:
                self._error_timeout *= 2

            logger.info("Could not connect to the stream, reconnection in %ss"
                        % self._error_timeout)

        elif self.state == ENHANCE_YOUR_CALM:
            if self._error_timeout < ENHANCE_YOUR_CALM_TIMEOUT:
                self._error_timeout = ENHANCE_YOUR_CALM_TIMEOUT
            else:
                self._error_timeout *= 2

            logger.warning("Enhance Your Calm response received from Twitter. "
                           "If you didn't restart your program frenetically "
                           "then there is probably something wrong with it. "
                           "Make sure you are not opening too many connections"
                           " to the endpoint you are currently using by "
                           "checking out Twitter's Streaming API "
                           "documentation: "
                           "https://dev.twitter.com/streaming/overview\n"
                           "The stream will restart in %ss."
                           % self._error_timeout)
        else:
            raise RuntimeError("Incorrect state: %d" % self.state)

        self._reconnecting = True
        return {'reconnecting_in': self._error_timeout, 'error': error}

    async def restart_stream(self):
        """
            Restart the stream on error
        """
        await self.response.release()
        await asyncio.sleep(self._error_timeout)
        await self.connect()

        logger.info("Reconnected to the stream")
        self._reconnecting = False
        return {'stream_restart': True}

    def __enter__(self):
        """
            Prepare the stream

        Returns
        -------
        StreamResponse
            The stream iterator
        """
        return self

    async def __aenter__(self):
        """
            Prepare the stream

        Returns
        -------
        StreamResponse
            The stream iterator
        """
        await self.client.setup
        return self

    def __exit__(self, *args):
        """ Close the response on error """
        if getattr(self, 'response', None) is not None:
            if not self.response.closed:
                logger.debug("Closing the stream")
                self.response.close()

    async def __aexit__(self, *args):
        """ Close the response on error """
        self.__exit__(*args)
